/**********************************************************************
Mini centrale DCC en une carte fille sur UNO ou MEGA
Equipée d'un écran Oled, d'un booster 6203, 3 boutons poussoirs et un encodeur quadratique 20 crans/tour
Recherche automatique de l'adresse d'une loco posée sur la voie principale lors de l'initialisation   
Version 107 du 5 Avril 2022
Ajout du mode de conduite manuelle pour tester locomotives et détecteurs

Affectation des Pins sur shield UNO
Pin 2   <-> ENC1 encodeur pin 1
Pin 3   <-> PWM (enable, HIGH = marche, LOW = stop) pour L6203
Pin 4   <-> ENC2 encodeur pin 2
Pin 5   <-> detecteur Gare 1
Pin 6   <-> detecteur zone A
Pin 7   <-> detecteur zone B
Pin 8   <-> detecteur zone C
Pin 9   <-> detecteur Gare 2
Pin 10  <-> DIR UNO pour L6203
Pin 11  <-> Led Rouge ERREUR
Pin 12  <-> DIR MEGA pour L6203
Pin 13  <-> bouton DOWN -> inter lumière 
Pin A0  <-> Mesure de courant L6203 via LM358
Pin A1  <-> bouton OK (encodeur)
Pin A2  <-> inter marche/arret -> bouton SEL  pin PIN_BtnSel
Pin A3  <-> bouton UP 
Pin A4  <-> SDA OLED
Pin A5  <-> SLC OLED
Pin A6  iibre non dispo sur UNO
Pin A7  libre non dispo sur UNO
***********************************************************/

const char Version[16] =  "V 108 03-05-22 "; // 15 caractères + \0
const int EEPVERSION = 108;

#include <Arduino.h>
#include <DCCpp.h> 
#include <EEPROM.h>
#include <Wire.h> 
#include <Bounce2.h>
#include "SSD1306Ascii.h"
#include "SSD1306AsciiAvrI2c.h"
#include <Encoder.h>            // encodeur

/******************
 * Constantes et Globales
 *****************/
#define DEBUG

#define  PIN_BtnEnc         A1    // poussoir OK (Encodeur)
#define  PIN_BtnSel         A2    // poussoir SEL (Marche / Arret DCC)
#define  PIN_BtnUp          A3    // poussoir UP (Manuel/Auto Automate)
#define  PIN_BtnDown        13    // poussoir DOWN (Lumière)
#define  PIN_GARE1           5     //capteur input
#define  PIN_ZONEA           6     //capteur input
#define  PIN_ZONEB           7     //capteur input
#define  PIN_ZONEC           8     //capteur input
#define  PIN_GARE2           9     //capteur input
#define  LEDROUGE           11    //led rouge output
#define  ENC1               2     //encoder input
#define  ENC2               4     //encoder input

//EEPROM
#define eeStart 0x30
struct ConfigEEprom {
  unsigned int vers;              // version logiciel
  unsigned long dist;             // = DISTANCE_MODULE; en cm
  int arret;                      // = ARRETENGARE;     en s
  int adresse;                    // = 3; ADRESSE par DEFAUT
  int vmin;                       // = vitesse minimum
};
ConfigEEprom gEPConfig;

//OLED
#define I2C_ADDRESS 0x3C // 0x3C or 0x3D
SSD1306AsciiAvrI2c oled;

//ENCODER
Encoder myEnc(2,4);             
long oldPosition  = -999;       // encoder
unsigned long encoderMillis;

//BOUTTONS
  Bounce BtnEnc = Bounce();
  Bounce BtnDown = Bounce();
  Bounce BtnUp = Bounce();
  Bounce BtnSel = Bounce();

//DCC
#ifndef USE_TEXTCOMMAND
#error To be able to compile this sample,the line #define USE_TEXTCOMMAND must be uncommented in DCCpp.h
#endif
int  gconsigne=40; 
bool gAutoMan=true;               // true en MARCHE automatique // false en Mode Manuel
bool gDCCStartStop=true;          // true DCC on // false DCC off
unsigned long gStartStopTime;     //
unsigned long gRepeatTime;        //
unsigned long gRepeatDelay = 500; //
unsigned long gCurrentSampleTime; //
byte gOccupation;                 // Etat des occupations (1 bit par zone)
bool gMesureVitesse;              //
unsigned long gcompteurVitesse;

#include "Automate.h"             // Trains, Zones, Automate (options pour va et vient)
                                  // gTrain créé avec adresse DCC 3
//BOOSTER
#define MOTOR_SHIELD_TYPE   0     // L6203 
#define COMM_INTERFACE      0
#ifdef ARDUINO_AVR_MEGA           // define as Mega 2560 (pinouts and functionality are identical)
#define ARDUINO_AVR_MEGA2560
#endif
#if defined(ARDUINO_AVR_UNO)
  #define PIN_DCC   10
#elif defined(ARDUINO_AVR_MEGA2560)
  #define PIN_DCC   12
#endif
#define PIN_ENABLE             3      // DCC_ENABLE / DISABLE
//COURANT/TENSION
#define CURRENT  A0               // current sensor
#define VOLTAGE  A2               // voltage sensor
int gcurrent=0;
const int CURRENTMAX = 400;
float gtension=0;

//TRAIN : gTrain declaré dans Automate.h


/////////////////////////////////////////////////////////////////////////////////////////
// RECHERCHE ADRESSE ET MISE EN ROUTE
/////////////////////////////////////////////////////////////////////////////////////////

void MiseEnRouteDCC() {
  int reponse = 3;
  oled.clear();
  oled.setCursor(0,0);
  oled.print("Lecture Adr");
  oled.setCursor(0,2);
  oled.print("DCC..");
  //EEPROM.get(eeStart, gEPConfig);           // lecture adresse en config
  start_DCC();
  gDCCStartStop = true;
  reponse = DCCpp::identifyLocoIdMain();
  Serial.print(F("adr="));Serial.print(reponse);
  if (reponse == -1) {
    oled.print("?");                        // 1er echec
    delay(500);
    reponse = DCCpp::identifyLocoIdMain();  // recommence
    if (reponse == -1) {
      oled.print("?");
      delay(500);
      reponse = gEPConfig.adresse; // prend l'adresse en config
    }
  } 
  gTrain.adresse_dcc = reponse;
  gEPConfig.adresse = reponse;
  EEPROM.put(eeStart, gEPConfig); // sauvegarde en config  
  oled.setCursor(0,2);
  oled.print("DCC =  ");
  oled.print(gTrain.adresse_dcc);
}

///////////////////////////////////////////////////////////////////////
// START / STOP DCC
///////////////////////////////////////////////////////////////////////

void start_DCC() {
  digitalWrite(PIN_ENABLE,HIGH);
  gDCCStartStop=true;
}

void stop_DCC() {
  digitalWrite(PIN_ENABLE,LOW);
  gDCCStartStop=false;
  delay(1000);
}

///////////////// Fonctions d'affichege OLED ///////////////////////
/* 11 caractères par ligne en taille X2
 *  Ligne 0 : DCC + adresse DCC + mode AUTO ou MAN
 *  Ligne 2 : Cran de vitesse + direction + lumière
 *  Ligne 3 : Nom Zone, vitesse en km/h ou decompte arrêt en gare
 *  Ligne 4 : Voltage et courant en mA
 *            
 * @dcc = -1 ou 24 ou 999
 * * = lumière
 * Lieu =  4 caractères
 * Cv = consigne en km/h
 * kmh = vitesse vraie en km/h
 */
/////// RAFRAICHISSEMENT DES AFFICHAGES PARAMETRES LOCO ///////////////////////

void oledCS(int n) {                            // compte de secondes
  if (n < 100) oled.print(" ");
  if (n < 10) oled.print(" ");
  oled.print(n);
}

void oledloco() {
  oled.setCursor(0,0);                          // 1ere ligne
  oled.print("@ ");
  oled.print(gTrain.adresse_dcc);
  oled.print(gAutoMan? " auto VV ":" manuel ");
  oled.print(gDCCStartStop? ' ':'!');
  oled.setCursor(0, 2);                         // 2eme ligne
  oled.print("Cran ");  
  oledCS(gTrain.cran_vitesse);
  oled.print(gTrain.sens?" >>":" <<");
  oled.print(gTrain.lumiere? " * ":"   ");
  oled.setCursor(0,4);                          // 3eme ligne
  oled.print(nomZone[gTrain.zone]);
  switch (gTrain.zone) {
    case ZGARE1:
    oled.print("Att ");oledCS(gTrain.compteur); oled.print("s  ");
    break;
    case ZGARE2:
    oled.print("Att ");oledCS(gTrain.compteur); oled.print("sec ");
    break;
    case ZONEA:
    oled.print("Vit ");oledCS(gTrain.vitesseVraie); oled.print("km/h ");
    break;
    case ZONEC:
    oled.print("Vit ");oledCS(gTrain.vitesseVraie); oled.print("km/h ");
    break;
    case ZONEB:
    oled.print("Csg ");oledCS(gTrain.cran_consigne); oled.print("km/h ");
    break;
    case INCONNU:
    oled.print("Csg ");oledCS(gTrain.cran_consigne); oled.print("km/h ");
    break;
  }
  oled.setCursor(0,6);                        // 4eme ligne
  oled.print((int)gtension);   
  oled.print(" Volts I=");
  oled.print(gcurrent*5);
  oled.print("mA ");
}

// A PLACER JUSTE AVANT SETUP POUR ACCEDER AUX GLOBALES
#include "Configuration.h"

//////////////////////////////////////////////////////////////////////////////////////////////////
void setup() {
  pinMode(LEDROUGE, OUTPUT);
  digitalWrite(LEDROUGE, HIGH); //allumé

  Serial.begin(115200);           // configure serial interface
  Serial.flush();
  Serial.println(Version);

  Wire.begin();         
  oled.begin(&Adafruit128x64, I2C_ADDRESS);
  //oled.setFont(Adafruit5x7);  
  oled.setFont(TimesNewRoman16_bold);
  //oled.setFont(Callibri15);
  oled.setLetterSpacing(2); // Increase space between letters.
  //oled.set2X();
  oled.clear();
  oled.setCursor(0,0);
  oled.print("Va et Vient");
  oled.setCursor(0,2);
  oled.print("Locoduino ");
  oled.setCursor(0,4);
  oled.print("Dominique ");
  oled.setCursor(0,6);
  oled.print("Vers ");
  oled.print(EEPVERSION);

  gtension = analogRead(VOLTAGE); // lecture tension alim
  gtension = gtension * 0.033;
  EEPROM.get(eeStart, gEPConfig);         // lecture configuration en EEPROM apres configuration
  /*#ifdef DEBUG
  Serial.print(F("version "));Serial.println(gEPConfig.vers);
  Serial.print(F("distance "));Serial.println(gEPConfig.dist);
  Serial.print(F("tempo arret "));Serial.println(gEPConfig.arret);
  Serial.print(F("addresse "));Serial.println(gEPConfig.adresse);
  Serial.print(F("V min "));Serial.println(gEPConfig.vmin);
  #endif*/
  delay(5000);              // appuyer sur le bouton de l'encoder pour modifier la configuration

  // Lecture de la configuration en EEPROM si appui sur bouton encodeur pendant le delay(5000)
  // necessite d'initialiser la pin
  pinMode(PIN_BtnEnc, INPUT_PULLUP);    // fait aussi par Bounce2
  int enc = digitalRead(PIN_BtnEnc);
  if (enc == LOW) {
    configuration();          // Configuration EEPROM lecture et ecriture si BEnc appuyé au demarrage: initialise gEPConfig
  }  

  BtnEnc.attach(PIN_BtnEnc, INPUT_PULLUP );
  BtnDown.attach(PIN_BtnDown, INPUT_PULLUP );
  BtnUp.attach( PIN_BtnUp, INPUT_PULLUP );
  BtnSel.attach( PIN_BtnSel, INPUT_PULLUP );

  BtnEnc.interval(25);
  BtnDown.interval(25);
  BtnUp.interval(25);
  BtnSel.interval(25);

  DCCpp::begin();
  DCCpp::beginMain(UNDEFINED_PIN, PIN_DCC, PIN_ENABLE, CURRENT);        // Dc: Dir, Pwm, current sensor for adress discovery
  MiseEnRouteDCC();             // et recherche de l'adresse de la loco ou en configuration si echec
    
  oldPosition = myEnc.read()/4;
    delay(1000);
  
  // init état LUMIERE
  
  gTrain.feux();                // lumiere OFF

  oledloco();
  digitalWrite(LEDROUGE, LOW);  //led éteinte 
  gzoneMonitor.Init();
  gTrain.cran_vitesse = gAutoMan? 15:0;     // vitesse 15 en Auto sinon stop
  gTrain.sens = true;
  gTrain.roule();
  gTrain.zone = INCONNU;
} // fin Setup

//////////////////////// Traitement boutons ///////////////////

void BtnUpPressed() {   // Manuel/Automate
  gAutoMan = !gAutoMan;
  if (gAutoMan) { // Auto
    gAutomate.LRclignote=true;
  } else { // Manuel
    gAutomate.LRclignote=false;
  }
  oledloco();           // mise a jour affichage
}

void BtnDownPressed() { // Lumière
  gTrain.lumiere = !gTrain.lumiere;
  gTrain.feux();        // selon lumiere ON ou OFF
  oledloco();           // mise a jour affichage
}

void BtnSelPressed() {  // Marche / Arret DCC
  gDCCStartStop = !gDCCStartStop;
  if (gDCCStartStop) {
    start_DCC(); 
  } else {
    stop_DCC(); 
    //gTrain.cran_vitesse = 0;
  }  
  Serial.println(gDCCStartStop? "DCC on":"DCC off");
  oledloco();           // mise a jour affichage
}

void BtnEncPressed() {  // Encodeur configuration ou stop train
  gTrain.stoppe();
  oledloco();           // mise a jour affichage
}

//////////////////////////////////////////////////////////////////////////////////////////////////
void loop() {
  DCCpp::loop();
  gAutomate.loop();
  oledloco();
  
  // Update button state
  BtnUp.update();
  if (BtnUp.fell()) { // manuel / auto 
    BtnUpPressed();
  }
  BtnDown.update();
  if (BtnDown.fell()) { // F0
    BtnDownPressed();
  }
  BtnSel.update();
  if (BtnSel.fell()) { // DCC start / stop
    BtnSelPressed();
  }
  BtnEnc.update();
  if (BtnEnc.fell()) { // stop
    BtnEncPressed();
  }  

  // affichage courant I = la valeur maxi pendant 2 secondes
  //int vc;
  if ((millis() - gCurrentSampleTime) > 5000) {
    gCurrentSampleTime = millis();
    gcurrent = 2; // raz toutes les secondes
    gtension = analogRead(VOLTAGE); // lecture tension alim
    gtension = gtension * 0.023;
  }
  
  // affichage courant et détection de court-circuit !!!
  int iCurrent = analogRead(CURRENT);
  if (iCurrent > gcurrent) {
    gcurrent = iCurrent;
    if (gcurrent > CURRENTMAX) { // 400 * 5 = 2000 mA
    stop_DCC();
    oled.setCursor(0,0);
    oled.print(F("C.CIRCUIT !"));
    delay(2000);
    } 
  }
  //////////////// ENCODEUR GESTION DE LA CONSIGNE DE VITESSE  ////////////////
  //////////////// EN MODE AUTO ET DE LA VITESSE EN MODE MANUEL //////////////
    
  long newPosition = myEnc.read()/4;
  bool delta;
  if (newPosition != oldPosition) {  
    delta = (newPosition > oldPosition); // > delta = true; < delta = false;
    #if defined(ARDUINO_AVR_UNO)
      delta = (newPosition > oldPosition); // > delta = true; < delta = false;
    #elif defined(ARDUINO_AVR_MEGA2560)
      delta = (newPosition < oldPosition); // > delta = true; < delta = false;
    #endif
    oldPosition = newPosition; 
    long intervalEncoder =  millis() - encoderMillis;
    encoderMillis = millis();
    int deltaCran = 5;
    if (intervalEncoder > 250) deltaCran = 1;
    if (!gAutoMan) {    // conduite manuelle 
      if (gTrain.sens) { 
        if (delta) {gTrain.cran_vitesse = gTrain.cran_vitesse + deltaCran;} else {gTrain.cran_vitesse = gTrain.cran_vitesse - deltaCran;}
      } else {
        if (delta) {gTrain.cran_vitesse = gTrain.cran_vitesse - deltaCran;} else {gTrain.cran_vitesse = gTrain.cran_vitesse + deltaCran;}
      }
      if (gTrain.cran_vitesse > 125) gTrain.cran_vitesse = 125;
      if (gTrain.cran_vitesse <= 0) {
        gTrain.sens = !gTrain.sens;
        gTrain.cran_vitesse = 0;
      } 
      gTrain.roule();
      oledloco();
    } else {      // consigne en mode auto
      if (delta) gconsigne++;  else  gconsigne--;     
      if (gconsigne < 0) gconsigne=0;
      if (gconsigne > 125) gconsigne=125;
      gTrain.cran_consigne = gconsigne;
      #ifdef DEBUG
      Serial.print("CONSIGNE");Serial.println(gTrain.cran_consigne);
      #endif
    }
  oledloco();           // mise a jour affichage 
  }
  
  ///////////////////// Détections de zones ///////////////////////////
  //////// Et initialisation de l'automate du train selon la zone /////
  
  if (gzoneMonitor.checkZone()) {     // toutes les 0,2 seconde
    Serial.println(nomZone[gTrain.zone]);
    switch (gOccupation) {                        // on ne traite qu'une nouvelle occupation
    case 1:      //ZGARE1
      gTrain.zone = ZGARE1;
      if (gAutoMan) {_doGare1();}
    break;
    case 2:      //ZONEA
      gTrain.zone = ZONEA;
      if (gAutoMan) {_doZoneA();}
    break;
    case 4:      //ZONEB
      gTrain.zone = ZONEB;
      if (gAutoMan) {_doZoneB();}
    break;
    case 8:      //ZONEC
      gTrain.zone = ZONEC;
      if (gAutoMan) {_doZoneC();}
    break;
    case 16:     //ZGARE2
      gTrain.zone = ZGARE2;
      if (gAutoMan) {_doGare2();}
    break;
    }
    oledloco();           // mise a jour affichage toutes les 0,1 s
  }
}
